"""
__version__.py
~~~~~~~~~~~~~~

Information about the current version of the py-package-template package.
"""

__title__ = 'groundhog'
__description__ = 'Groundhog - A general-purpose geotechnical package'
__version__ = '0.1.1'
__author__ = 'Bruno Stuyts'
__author_email__ = 'bruno@pro-found.be'
__license__ = 'Creative Commons BY-SA 4.0'
__url__ = 'https://github.com/snakesonabrain/groundhog'
__download_url__ = 'https://github.com/snakesonabrain/groundhog/archive/master.zip'
